﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO.Ports;
using System.Linq;
using System.Net.Http;
using System.Security.Permissions;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using TelemetryIO.Models;

namespace TelemetryIO
{
    public class SerialHandler
    {
        public const int TELE_CMD_RD_ONCE =   1;
        private const int TELE_CMD_RD_MON_ON =   2;
        private const int TELE_CMD_RD_MON_OFF = 3;
        private const int TELE_CMD_RD_MON_ADD = 4;
        private const int TELE_CMD_RD_MON_REMOVE = 5;
        private const int TELE_CMD_RD_MON_REMOVEALL = 6;
        private const int TELE_CMD_WR =     10;
        private const int TELE_CMD_MOTORS_CTRL = 100;
        private const int TELE_CMD_ABORT =     999;
        private const int TELE_CMD_HELLO =     9999;
        private const byte CRC8_POLYNOMIAL = 0x07;
        private const int CRC32_POLY = 0x04C11DB7;
        private const byte ESCAPE_BEGIN = 0xBE;
        private const byte ESCAPE_END = 0xED;
        private const byte ESCAPE_CHAR = 0x0E;
        private SerialPort com_port;
        static private List<byte> rx_buf = new List<byte>();
        private byte[] saving_request = new byte[128];

        //Generate event when answer is received
        public delegate void AnswerEventHandler<SerialEventArgs>(object sender, SerialEventArgs e);
        public event AnswerEventHandler<SerialEventArgs> AnswerReceived;//событие получены данные в ответ на запрос

        //Generate event when handshake is occurred
        public delegate void HandShakeHandler(object sender, EventArgs e);
        public event HandShakeHandler HandShakeOccurred;//событие полетник ответил на запрос HELLO

        //Generate event when monitoring telegram is received
        public delegate void MonitoringEventHandler<MonitoringEventArgs>(object sender, MonitoringEventArgs e);
        public event MonitoringEventHandler<MonitoringEventArgs> MonitoringItemsReceived;//событие получены данные мониторинга

        private int int_test = 0;
        private float[] monitor = new float[32];
        public string view_str = "";
        Telemetry telemetry = Telemetry.Instance;

        private int currentCmd = 0;

        private Queue<byte[]> req_buffer = new Queue<byte[]>();
        private object lock_obj = new object();
        private object lock_obj_put = new object();
        private bool waitingForResponse = false;
        private bool isMonitorEmpty = false;
        private bool isTimeout = false;
        private readonly int responseTimeout = 2000;
        private System.Timers.Timer timeout = new System.Timers.Timer(3000);

        private bool exitPending = false;
        

        public struct Telegram
        {
            public int cmd;
            public int slot;
            public int len;
            public int offset;
            public byte[] param;
        };

        public string PortName
        {
            get { return com_port.PortName; }
            set { com_port.PortName = value; }
        }

        public int BaudRate
        {
            get { return com_port.BaudRate; }
            set { com_port.BaudRate = value; }
        }

        public bool isOpen
        {
            get { return com_port.IsOpen; }
        }

        public SerialHandler()
        {
            com_port = new SerialPort();
            timeout.AutoReset = false;
        }

        public SerialHandler(string PortName, int BaudRate)
        {
            com_port = new SerialPort();
            com_port.PortName = PortName;
            com_port.BaudRate = BaudRate;
            timeout.AutoReset = false;
        }

        private void Serial_DataReceived(object sender, SerialDataReceivedEventArgs e)
        {
            //Прием данных при получении ESCAPE_END считаем, что пакет полностью получен
            //и его можно парсить
            //СТРУКТУРА ПАКЕТА
            //КОМАНДА[4 байта], АДРЕСС ЯЧЕЙКИ[4 байта], ДЛИНА ДАННЫХ[4 байта], СМЕЩЕНИЕ[4 байта], ДАННЫЕ[в пределах размера буффера], CRC32[4 байта]
            SerialPort sp = (SerialPort)sender;
            try
            {
                while (sp.BytesToRead > 0)
                {
                    if (!sp.IsOpen) return;
                    byte b = (byte)sp.ReadByte();

                    if (b == ESCAPE_END)
                    {//END BYTE IS RECEIVED
                        isTimeout = false;
                        byte[] unscape = EscapeSeqToBytes(rx_buf.ToArray());
                        uint checksum = crc32(unscape, unscape.Length - 4);
                        uint re_checksum = DataSlots.bytes_to_uint(unscape, unscape.Length - 4);
                        if (re_checksum == checksum)
                        {
                            //Parse telegram
                            int cmd = DataSlots.bytes_to_int(unscape, 0);
                            int slot = DataSlots.bytes_to_int(unscape, 4);
                            int len = DataSlots.bytes_to_int(unscape, 8);
                            int offset = DataSlots.bytes_to_int(unscape, 12);
                            byte[] data = new byte[len];
                            Array.Copy(unscape, 16, data, 0, len);
                        Debug.WriteLine($"cmd = {cmd} *** slot = {slot} *** len = {len}");

                        if (cmd == TELE_CMD_RD_ONCE)
                            {
                                telemetry.setSlot(slot, data, offset, len);
                                OnAnswerReceived(slot.ToString());
                                //DataSlots.setSlot(slot, data, offset, len);
                            }
                            else if (cmd == TELE_CMD_RD_MON_ON)
                            {//READ MONITORING DATA
                                for (int i = 0; i < data.Length / 4; i++)
                                {
                                    monitor[i] = BitConverter.ToSingle(data, i * 4);
                                }
                                OnMonitoringItemsReceived(monitor);
                                OnAnswerReceived("MONITORON");
                                view_str = string.Join(", ", monitor.Select(x => "" + x).ToArray(), 0, data.Length / 4);
                            }
                            else if (cmd == TELE_CMD_WR)
                            {//WRITTING OPERATION IS DONE WITH SUCCESS
                                OnAnswerReceived("Data is written");
                            }
                            else if (cmd == TELE_CMD_MOTORS_CTRL)
                            {//WRITTING OPERATION IS DONE WITH SUCCESS
                                //Debug.WriteLine($"cmd = {cmd} *** slot = {slot} *** len = {len}");
                                telemetry.setSlot(Telemetry.MOTORS_ACT_ADDRESS, data, offset, len);
                                OnAnswerReceived("MOTORSACT");

                            }
                            else if (cmd == TELE_CMD_HELLO)
                            {//HANDSHAKING ANSWER
                                telemetry.setSlot(9999, data, 0, 0);
                                view_str = Encoding.ASCII.GetString(data);
                                req_buffer.Clear();
                                OnHandShakeOccurred();
                            }
                            else if (cmd == TELE_CMD_RD_MON_ADD)
                            {
                                OnAnswerReceived("ADD");
                                Debug.WriteLine($"cmd = {cmd} *** slot = {slot} *** len = {len}");
                                isMonitorEmpty = false;
                            }
                            else if (cmd == TELE_CMD_RD_MON_REMOVE)
                            {
                                OnAnswerReceived("REMOVE");
                                Debug.WriteLine($"cmd = {cmd} *** slot = {slot} *** len = {len}");
                            }
                            else if (cmd == TELE_CMD_RD_MON_REMOVEALL)
                            {
                                if (exitPending)
                                {
                                    com_port.Close();
                                    OnAnswerReceived("EXITAPP");
                                }
                                else OnAnswerReceived("REMOVEALL");
                            }
                            else if (cmd == TELE_CMD_RD_MON_OFF)
                            {
                                OnAnswerReceived("MONITOROFF");
                            }
                            
                            waitingForResponse = false;
                            timeout.Stop();
                            sendNextRequest();
                        }
                    }
                    else if (b == ESCAPE_BEGIN)
                    {//START BYTE IS RECEIVED
                        rx_buf.Clear();
                    }
                    else
                    {//FILLING BUFFER
                        rx_buf.Add(b);
                    }
                }
            }catch(Exception ex)
            {
                OnAnswerReceived("EXCEPTION" + ex.Message);
            }
        }

        public void Open()
        {
            com_port.Open();
            com_port.DataReceived += Serial_DataReceived;
            timeout.Elapsed += Timeout_Elapsed;
            putRequest(prepareTelegram(TELE_CMD_HELLO, 0, new byte[0], 0));
        }

        private void Timeout_Elapsed(object sender, System.Timers.ElapsedEventArgs e)
        {
            //Таймаут коммуникации. Без проверки на правильность ответа
            System.Timers.Timer t = sender as System.Timers.Timer;
            
            lock(lock_obj)
            {
                if (waitingForResponse)
                {
                    //generate timeout event
                    Debug.WriteLine("Timeout!!!!!");
                    waitingForResponse = false;
                    isTimeout = true;
                    if(exitPending) OnAnswerReceived("EXITAPP");
                    else OnAnswerReceived("TIMEOUT");
                    req_buffer.Clear();
                    //this.Close();
                }
            }
        }

        public void Close()
        {
            com_port.DataReceived -= Serial_DataReceived;
            timeout.Elapsed -= Timeout_Elapsed;
            com_port.Close();
        }

        public void requestExit()
        {
            //запрос на закрытие приложения, отправляем команду на очистку массива мониторинга и ждем ответ, либо таймаут,
            //после чего приложение закрывается
            RemoveMonitoringItems();
            exitPending = true;
        }

        public void putRequest(byte[] request)
        {
            //добавить в очередь или отправить(если очередь пуста) пакет в порт
            lock(lock_obj_put)
            {
                if (waitingForResponse)
                {
                    req_buffer.Enqueue(request);
                    return;
                }

                sendRequest(request);
            }
        }

        private void sendRequest(byte[] request)
        {
            //отправка пакета
            waitingForResponse = true;
            if(com_port.IsOpen)com_port.Write(request, 0, request.Length);
            timeout.Stop();
            timeout.Start();
        }

        private void sendNextRequest()
        {
            //вытащить из очереди пакет и отправить в порт
            waitingForResponse = false;
            if (req_buffer.Count > 0)
            {
                sendRequest(req_buffer.Dequeue());
            }
        }

        //**********************************************************************************************
        //*************************************** REQUESTS BLOCK ***************************************
        //**********************************************************************************************
        public void getPIDs()
        {
            //отправить наборы ПИДов
            putRequest(prepareTelegram(TELE_CMD_RD_ONCE, 1000, new byte[0], 0, 4 * 9));
            putRequest(prepareTelegram(TELE_CMD_RD_ONCE, 1001, new byte[0], 0, 4 * 9));
            putRequest(prepareTelegram(TELE_CMD_RD_ONCE, 1002, new byte[0], 0, 4 * 9));
            putRequest(prepareTelegram(TELE_CMD_RD_ONCE, 1003, new byte[0], 0, 4 * 9));
        }

        public void stopMonitoring()
        {
            //остановить мониторинг
            putRequest(prepareTelegram(TELE_CMD_RD_MON_OFF, 0, new byte[0], 0, 0));
        }

        public void startMonitoring()
        {
            //начать мониторинг
            putRequest(prepareTelegram(TELE_CMD_RD_MON_ON, 0, new byte[0], 0, 0));
        }

        public void AddMonitoringItem(int slot, int offset)
        {
            //добавить элемент из массива мониторинга по адресу
            putRequest(prepareTelegram(TELE_CMD_RD_MON_ADD, slot, new byte[0], offset));
        }

        public void AddMonitoringItem(string name)
        {
            //добавить элемент из массива мониторинга по имени
            putRequest(prepareTelegram(TELE_CMD_RD_MON_ADD, new byte[0], name));
        }

        public void RemoveMonitoringItem(int position)
        {
            //удалить элемент из массива мониторинга
            putRequest(prepareTelegram(TELE_CMD_RD_MON_REMOVE, position, new byte[0], 0));
        }

        public void RemoveMonitoringItems()
        {
            //удалить все элементы из массива мониторинга
            putRequest(prepareTelegram(TELE_CMD_RD_MON_REMOVEALL, 0, new byte[0], 0));
        }

        public void sendFloats(int slot, float[] sp)
        {
            //Записать массив чисел с плавающей точкой
            putRequest(prepareTelegram(TELE_CMD_WR, slot, sp, 0, sp.Length*4));
        }

        public void sendMotorsControl()
        {
            //Отправляет задание на моторы в полетник. Тот в ответ посылает актуальные скорости на моторах
            //В offset передается количество моторов
            putRequest(prepareTelegram(TELE_CMD_MOTORS_CTRL, Telemetry.MOTORS_SP_ADDRESS, telemetry.motor_sp, 8));
        }

        //**********************************************************************************************
        //**********************************************************************************************
        //**********************************************************************************************


        public void saveFloats(float[] data)
        {
            prepareTelegram(TELE_CMD_WR, 1000, data, 0);
        }

        byte crc8(byte[] data, int length)
        {
            byte crc = 0x00;  // Initial value

            for (int i = 0; i < length; i++)
            {
                crc ^= data[i];  // XOR with the data byte

                for (int j = 0; j < 8; j++)
                {
                    if ((byte)(crc & 0x80) > 0)
                    {
                        crc = (byte)((byte)(crc << 1) ^ CRC8_POLYNOMIAL);
                    }
                    else
                    {
                        crc = (byte)(crc << 1);
                    }
                }
            }

            return crc;
        }

        uint crc32(byte[] data, int length)
        {
            uint crc = 0xFFFFFFFF;  // Начальное значение CRC

            for (int i = 0; i < length; i++)
            {
                crc ^= (uint)data[i] << 24;  // XOR с текущим байтом

                for (int j = 0; j < 8; j++)
                {
                    if ((uint)(crc & 0x80000000) != 0)
                    {
                        crc = (crc << 1) ^ CRC32_POLY;
                    }
                    else
                    {
                        crc <<= 1;
                    }
                }
            }

            return crc;
        }
        byte[] prepareTelegram<T>(int cmd, T[] load, string var_name)
        {
            VarAddress va = telemetry.getVarAdress(var_name);
            return prepareTelegram(cmd, va.slot, load, va.offset, va.length);
        }

        byte[] prepareTelegram<T>(int cmd, int slot, T[] load, int offset, int len = 0)
        {
            byte[] byteload = DataArrayToBytes(load);
            int total_len = 20 + byteload.Length;//cmd[4 bytes] + slot[4 bytes] + len[4 bytes] + offset[4 bytes] + load[len bytes]
            byte[] data = new byte[total_len];

            //Construct telegram
            data[0] = (byte)(0xFF & cmd);
            data[1] = (byte)(0xFF & (cmd >> 8));
            data[2] = (byte)(0xFF & (cmd >> 16));
            data[3] = (byte)(0xFF & (cmd >> 24));

            data[4] = (byte)(0xFF & slot);
            data[5] = (byte)(0xFF & (slot >> 8));
            data[6] = (byte)(0xFF & (slot >> 16));
            data[7] = (byte)(0xFF & (slot >> 24));

            int l = 0;
            if (cmd == TELE_CMD_WR) l = byteload.Length;
            else l = len;

            data[8] = (byte)(0xFF & l);
            data[9] = (byte)(0xFF & (l >> 8));
            data[10] = (byte)(0xFF & (l >> 16));
            data[11] = (byte)(0xFF & (l >> 24));

            data[12] = (byte)(0xFF & offset);
            data[13] = (byte)(0xFF & (offset >> 8));
            data[14] = (byte)(0xFF & (offset >> 16));
            data[15] = (byte)(0xFF & (offset >> 24));

            if(byteload.Length > 0)
            {
                //Copy data
                Array.Copy(byteload, 0, data, 16, byteload.Length);
            }
            //CRC32
            uint checksum = crc32(data, total_len - 4);
            data[total_len - 4] = (byte)(0xFF & checksum);
            data[total_len - 3] = (byte)(0xFF & (checksum >> 8));
            data[total_len - 2] = (byte)(0xFF & (checksum >> 16));
            data[total_len - 1] = (byte)(0xFF & (checksum >> 24));

            byte[] escape = BytesToEscapeSeq(data);
            byte[] ret = new byte[escape.Length + 2];
            Array.Copy(escape, 0, ret, 1, escape.Length);
            ret[0] = ESCAPE_BEGIN;
            ret[ret.Length - 1] = ESCAPE_END;
            Array.Copy(ret, saving_request, ret.Length);
            return ret;
        }

        byte[] DataArrayToBytes<T>(T[] data)
        {
            if (data == null) return new byte[0];
            List<byte>ret = new List<byte> ();
            for(int i = 0; i < data.Length; i++)
            {
                if(typeof(T) == typeof(float))
                {
                    ret.AddRange(BitConverter.GetBytes((float)(object)data[i]));
                }else if(typeof(T) == typeof(int))
                {
                    ret.AddRange(BitConverter.GetBytes((int)(object)data[i]));
                }else if(typeof(T) == typeof(byte))
                {
                    ret.Add((byte)(object)data[i]);
                }else if(typeof(T) == typeof(bool))
                {
                    bool t = (bool)(object)data[i];
                    if (t) ret.Add(1);
                    else ret.Add(0);
                }
                
            }
            return ret.ToArray();
        }

        byte[] BytesToEscapeSeq(byte[] data)
        {
            List<byte> ret = new List<byte>();
            for (int i = 0; i < data.Length; i++)
            {
                if ((data[i] == ESCAPE_BEGIN) || (data[i] == ESCAPE_CHAR) || (data[i] == ESCAPE_END))
                {
                    ret.Add(ESCAPE_CHAR);
                    ret.Add((byte)(data[i] - 0x0E));
                }
                else
                {
                    ret.Add(data[i]);
                }
            }
            return ret.ToArray();
        }

        byte[] EscapeSeqToBytes(byte[] EscSeq)
        {
            List<byte> ret = new List<byte>();
            for (int i = 0; i < EscSeq.Length; i++)
            {
                //if ((EscSeq[i] == ESCAPE_BEGIN) || (EscSeq[i] == ESCAPE_CHAR) || (EscSeq[i] == ESCAPE_END))
                if (EscSeq[i] == ESCAPE_CHAR)
                {
                    i++;
                    ret.Add((byte)(EscSeq[i] + 0x0E));
                }
                else
                {
                    ret.Add(EscSeq[i]);
                }
            }
            return ret.ToArray();
        }

        protected virtual void OnAnswerReceived(string answer)
        {
            AnswerReceived?.Invoke(this, new SerialEventArgs(answer));
        }

        protected virtual void OnHandShakeOccurred()
        {
            HandShakeOccurred?.Invoke(this, new EventArgs());
        }

        protected virtual void OnMonitoringItemsReceived(float[] data)
        {
            MonitoringItemsReceived?.Invoke(this, new MonitoringEventArgs(data));
        }
    }

    public class SerialEventArgs: EventArgs
    {
        public string Answer { get; set; }

        public SerialEventArgs(string answer)
        {
            Answer = answer;
        }
    }

    public class MonitoringEventArgs: EventArgs
    {
        public float[] Data { get; set;}

        public MonitoringEventArgs(float[] data)
        {
            Data = data;
        }
    }
}
